@@ -10,6 +10,7 @@
 #include <unordered_map>
 
 #include <QDeadlineTimer>
+#include <QElapsedTimer>
 #include <QQmlComponent>
 #include <QTimer>
 
@@ -60,6 +61,7 @@ public:
     QList<QObject *> hiddenActions;
     QQmlComponent *fullDelegate = nullptr;
     QQmlComponent *iconDelegate = nullptr;
+    QQmlComponent *separatorDelegate = nullptr;
     QQmlComponent *moreButton = nullptr;
     qreal spacing = 0.0;
     Qt::Alignment alignment = Qt::AlignLeft;
@@ -219,6 +221,23 @@ void ToolBarLayout::setIconDelegate(QQmlComponent *newIconDelegate)
     Q_EMIT iconDelegateChanged();
 }
 
+QQmlComponent *ToolBarLayout::separatorDelegate() const
+{
+    return d->separatorDelegate;
+}
+
+void ToolBarLayout::setSeparatorDelegate(QQmlComponent *newSeparatorDelegate)
+{
+    if (newSeparatorDelegate == d->separatorDelegate) {
+        return;
+    }
+
+    d->separatorDelegate = newSeparatorDelegate;
+    d->delegates.clear();
+    relayout();
+    Q_EMIT separatorDelegateChanged();
+}
+
 QQmlComponent *ToolBarLayout::moreButton() const
 {
     return d->moreButton;
@@ -329,11 +348,7 @@ void ToolBarLayout::componentComplete()
 void ToolBarLayout::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
 {
     if (newGeometry != oldGeometry) {
-        if (newGeometry.size() != QSizeF{implicitWidth(), implicitHeight()}) {
-            relayout();
-        } else {
-            polish();
-        }
+        relayout();
     }
     QQuickItem::geometryChange(newGeometry, oldGeometry);
 }
@@ -371,7 +386,7 @@ void ToolBarLayoutPrivate::calculateImplicitSize()
         return;
     }
 
-    if (!fullDelegate || !iconDelegate || !moreButton) {
+    if (!fullDelegate || !iconDelegate || !separatorDelegate || !moreButton) {
         qCWarning(KirigamiLayoutsLog) << "ToolBarLayout: Unable to layout, required properties are not set";
         return;
     }
@@ -426,6 +441,8 @@ void ToolBarLayoutPrivate::calculateImplicitSize()
 
     visibleActionsWidth = 0.0;
 
+    q->setImplicitWidth(maxWidth);
+
     if (maxWidth > q->width() - (hiddenActions.isEmpty() ? 0.0 : moreButtonInstance->width() + spacing)) {
         // We have more items than fit into the view, so start hiding some.
 
@@ -458,7 +475,8 @@ void ToolBarLayoutPrivate::calculateImplicitSize()
         maxHeight = std::max(maxHeight, moreButtonInstance->implicitHeight());
     };
 
-    q->setImplicitSize(maxWidth, maxHeight);
+    q->setImplicitHeight(maxHeight);
+
     Q_EMIT q->hiddenActionsChanged();
 
     implicitSizeValid = true;
@@ -622,12 +640,26 @@ ToolBarLayoutDelegate *ToolBarLayoutPrivate::createDelegate(QObject *action)
         fullComponent = fullDelegate;
     }
 
+    auto separator = action->property("separator");
+    if (separator.isValid() && separator.toBool()) {
+        fullComponent = separatorDelegate;
+    }
+
     auto result = new ToolBarLayoutDelegate(q);
     result->setAction(action);
     result->createItems(fullComponent, iconDelegate, [this, action](QQuickItem *newItem) {
         newItem->setParentItem(q);
         auto attached = static_cast<ToolBarLayoutAttached *>(qmlAttachedPropertiesObject<ToolBarLayout>(newItem, true));
         attached->setAction(action);
+
+        if (!q->childItems().isEmpty() && q->childItems().first() != newItem) {
+            // Due to asynchronous item creation, we end up creating the last item
+            // first. So move items before previously inserted items to ensure
+            // we have a more sensible tab order.
+            // Note that this will be incorrect if we end up completing in random
+            // order.
+            newItem->stackBefore(q->childItems().first());
+        }
     });
 
     return result;
